# MULTRIS ARTIFACT - COQ DEVELOPMENT

This is the Coq mechanisation of the Multris program logic presented in the OOPSLA'24 paper

`Multris: Functional Verification of Multiparty Message Passing in Separation Logic`

## Multris Overview

Multris is a program logic for verifying safety (crash-freedom)
and functional correctness (e.g. in the presence of asserts) of
multiparty message-passing programs,
where messages exchanged between actors depends on each other.

An example of such a program is:
```
roundtrip_prog :=
   let (c0,c1,c2) := new_chan #3 in
   Fork (let x := recv c1 0 in send c1 2 (x+20));
   Fork (let x := recv c2 1 in send c2 0 (x+30));
   send c0 1 100; let x := recv c0 2; assert (x = 150).
```
Here `c0` sends 100 to `c1`, and expects to receive 150 (100 + 50) from `c2`.

Multris yields separation logic rules in the style of multiparty session types,
for verifying the _consistency_ of network protocols (when resolving `new_chan` instructions),
and for validating the sound _conformance_ to them (when resolving `send`/`recv` instructions).
In particular, the protocol for `c0` can be defined as:
```
c0 ↣ ![1] (x:Z) <x> . ?[2] <x+50> . end
```
This denotes that `c0` must first send a freely picked integer `x` to participant `1`,
after which point it can receive `x + 50` from participant `2`.

The Multris protocols and rules can be used to prove Hoare triples like the one for roundtrip prog:
```
{ True } roundtrip_prog { True }
```
Top-level Hoare triples (with no preconditions) yield safety and functional correctness
via the foundationally proven (in Coq) adequacy theorem of Multris:
```
if { True } e { True} then
e is safe to execute
```

This Coq mechanisation fully evidence each of the claimed contributions of the paper, namely:

1. Protocol consistency in separation logic
2. Tactical support for protocol consistency
3. Foundational verification (of multiparty message-passing programs)
4. (Verification of) Multiparty verification benchmark (and all other examples in the paper)

In particular, the mechanisation:

1. Provides a concrete definition and meta-theory for protocol consistency in Iris/Coq.
2. Provides Coq tactics for proving individual protocol consistency definitions
3. Defines and validates the program logic rules presented in the paper, that readily connects with the Iris adequacy theorem
4. Defines and verifies the Chang and Roberts ring leader benchmark and all other examples presented in the paper

In this README we cover:

- The hardware dependencies of the artifact
- How to get started and install and use the Multris mechanisation
- A step-by-step instruction on how the artifact validates the above claims
- A reusability guide for people who wants to use Multris

## Hardware Dependencies

The artifact has no specific hardware dependencies.
It has been tested on a MacBook Pro, 13-inch, M1, 2020 machine.
The artifact consists of the Coq source code of the mechanisation,
which can be built and executed even on low-end hardware.

## Getting Started Guide

The Coq mechanisation artifact has been built and tested with the following dependencies:

 - OCaml 4.12
 - Coq 8.19.0
 - The development version of Iris with git hash [c1e15cdc]

The depedencies can be installed with opam version 2.1.5, which is available
through most package managers: https://opam.ocaml.org/doc/Install.html

To avoid interference with existing local installations in opam, it is
recommended to make a new switch with
`opam switch create multris ocaml-base-compiler.4.12.0`

The Coq dependency can be installed via
`opam pin add coq 8.19.0`

The Iris dependency can then be installed via
`opam repo add iris-dev https://gitlab.mpi-sws.org/iris/opam.git`
followed by
`opam install coq-iris-heap-lang.dev.2024-03-12.0.c1e15cdc`

In order to build the artifact, install the above dependencies and then run
`make -j [num CPU cores]` to compile Multris.

To use Multris we recommend that the user has installed a Coq IDE, such as
CoqIDE, Emacs with ProofGeneral, or Visual Studio Code with vscoq.
These allow users to step through the Coq mechanisation interactively.

Using Multris to verify a multiparty message-passing program usually consists of
three steps.

1. Define the network of protocols
2. Prove consistency of the network using the Multris consistency tactics
3. Verify the program using the Iris/Multris symbolic execution tactics

As an initial test, we recommend the user to step through the proof of `roundtrip_prog`
in [multi_actris/examples/basics.v](multi_actris/examples/basics.v) (until line 54).

We defer additional details on the individual Coq tactics/constructions to the reusability guide.

We produced the line counts of Table 1 in the paper using
`wc $(find ./ -name \*.v)`

## Step-by-Step Instructions

In this section we go through each of the contributions claimed by the paper.

We present the claims in the order of their depedencies within the mechanisation: 1->3->2->4.

We recommend the user to check the correspondences between the definitions and rules between the
paper and the mechanisation.

We provide tables that explicitly show the direct correspondence between any such definitions and rules.

We start out by remarking some differences between the paper and the mechanisation.

### Differences between the paper and mechanisation

The following table gives a mapping between the notation in the paper and the Coq mechanisation:

Multiparty Dependent Separation Protocols:

|        | OOPSLA'24 paper                  | Coq mechanization                              |
|--------|----------------------------------|------------------------------------------------|
| Send   | `![i] x_1 .. x_n <v>{ P }. prot` | `<(Send,i) @ x_1 .. x_n> MSG v {{ P }}; prot`  |
| Recv   | `?[i] x_1 .. x_n <v>{ P }. prot` | `<(Recv,i) @ x_1 .. x_n> MSG v {{ P }}; prot`  |
| End    | `end`                            | `END`                                          |

The protocol for `c0` above is thus written as:
```
c0 ↣ <(Send, 1) @ (x:Z)> MSG #x ; <(Recv, 2)> MSG #(x+50); END
```

The creation of channels in the Coq mechanisation projects each of the channel endpoints out explicitly,
after channel network creation, via a procedure `get_chan`.
As such, we also have two separate rules for `new_chan` and `get_chan`, that together yields the behaviour
of `wp_new` of the paper.
Even so, we have a unified tactic for resolving the creation of channel networks alongside projecting it
into its individual channel endpoints (as described in the reusability guide).

Finally, the Coq mechanisation use custom syntax for writing programs, inherited from Iris.
As an example the roundtrip program example from above is written as:
```
Definition roundtrip_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork (let: "x" := recv "c1" #0 in send "c1" #2 ("x"+#20));;
     Fork (let: "x" := recv "c2" #1 in send "c2" #0 ("x"+#30));;
     send "c0" #1 #100;; recv "c0" #2.
```
We refer to the Iris documentation for additional details:
https://gitlab.mpi-sws.org/iris/iris/-/blob/master/docs/heap_lang.md

### Claim 1: Protocol consistency in separation logic

The protocol consistency and the meta-theory behind it is found in the
[multi_actris/channel/proto.v](multi_actris/channel/proto.v) file.

An overview of the correspondences between the most significant definitions and rules of paper and the mechanisation is as follows:

| Paper                                              | Mechanisation                                                                      |
-----------------------------------------------------|------------------------------------------------------------------------------------|
| Multiparty Dependent Separation Protocols          | [multi_actris/channel/proto.v](multi_actris/channel/proto.v)                       |
| - `iProto` (pg. 20)                                | `iProto`                                                                           |
| - `end` (pg. 20)                                   | `iProto_end`                                                                       |
| - `send/recv` (pg. 20)                             | `iProto_message`                                                                   |
| Protocol Consistency                               | [multi_actris/channel/proto.v](multi_actris/channel/proto.v)                       |
| - `dual` definition (pg. 20)                       | `can_step`                                                                         |
| - `present` definition (pg. 20)                    | `valid_target`                                                                     |
| - `consistent` efinition (pg. 20)                  | `iProto_consistent`                                                                |
| - Rules (fig. 5)                                   | Directly from definition                                                           |
| Multris Ghost Theory                               | [multi_actris/channel/proto.v](multi_actris/channel/proto.v)                       |
| - `proto-alloc` rule (fig. 7)                      | `iProto_init`                                                                      |
| - `proto-step` rule (fig. 7)                       | `iProto_step`                                                                      |
| - `prot_ctx` fragment (pg. 20)                     | `iProto_ctx`                                                                       |
| - `prot_own` fragment (pg. 20)                     | `iProto_own`                                                                       |

### Claim 3: Foundational verification (of multiparty message-passing programs)

The shallow embedding of our multiparty channel operations and their specifications can be found in the
[multi_actris/channel/channel.v](multi_actris/channel/channel.v) file.

An overview of the correspondences between the most significant definitions and rules of paper and the mechanisation is as follows:

| Paper                                              | Mechanisation                                                                      |
-----------------------------------------------------|------------------------------------------------------------------------------------|
| Channel Implementation                             | [multi_actris/channel/channel.v](multi_actris/channel/channel.v)                   |
| - `new_chan` implementation (fig. 6)               | `new_chan` / `get_chan`                                                            |
| - `send` implementation (fig. 6)                   | `send`                                                                             |
| - `recv` implementation (fig. 6)                   | `recv`                                                                             |
| Channel Endpoint Ownership                         | [multi_actris/channel/channel.v](multi_actris/channel/channel.v)                   |
| - `c ↣ prot` (fig. 8)                              | `iProto_pointsto`                                                                  |
| - `chan_inv` (fig. 8)                              | `chan_inv`                                                                         |
| Channel Rules                                      | [multi_actris/channel/channel.v](multi_actris/channel/channel.v)                   |
| - `new_chan` rule [wp-new] (fig. 3)                | `new_chan_spec` + `get_chan_spec`                                                  |
| - `send` rule [wp-send] (fig. 3)                   | `send_spec`                                                                        |
| - `recv` rule [wp-recv] (fig. 3)                   | `recv_spec`                                                                        |
| Adequacy theorem (pg. 16)                          | Directly inherited from Iris, via shallow embedding                                |

### Claim 2: Tactical support for protocol consistency

The tactics for protocol consistency (and for resolving the channel constructors `new_chan`, `send`, and `recv`) can be found in the
[multi_actris/channel/proofmode.v](multi_actris/channel/proofmode.v) file.

The protocol consistency tactic referred to by the paper is `iProto_consistent_take_steps`.
While parsing it requires advanced Coq/Iris understanding, we want to emphasise how it is
constructed strictly using surface-level Iris Proof Mode (IPM) and Coq tactics,
and leverage features of IPM such as resource framing (via `iFrame`).

We will see how it is practically applied in reusability guide.

### Claim 4: (Verification of) Multiparty verification benchmark (and all other examples in the paper)

The examples of the paper can be found in the following files:

- Basic example [multi_actris/examples/basics.v](multi_actris/examples/basics.v)
- Chang and Roberts ring leader election [multi_actris/examples/leader_election.v](multi_actris/examples/leader_election.v)
- Two-buyer protocol [multi_actris/examples/two_buyer.v](multi_actris/examples/two_buyer.v)
- Three-buyer protocol [multi_actris/examples/three_buyer.v](multi_actris/examples/three_buyer.v)

An overview of the correspondences between the most significant definitions and rules of paper and the artifact is as follows:

| Paper                                              | Mechanisation                                                                      |
-----------------------------------------------------|------------------------------------------------------------------------------------|
| Basic Examples                                     | [multi_actris/examples/basics.v](multi_actris/examples/basics.v)                   |
| - Motivating program (pg. 2)                       | smuggle_ref_prog                                                                   |
| - Motivating program protocol (pg. 2/10)           | iProto_smuggle_ref                                                                 |
| - Roundtrip program (pg. 4)                        | roundtrip_prog                                                                     |
| - Roundtrip protocol (pg. 4)                       | iProto_roundtrip                                                                   |
| - Roundtrip reference program (pg. 5)              | roundtrip_ref_prog                                                                 |
| - Roundtrip reference protocol (pg. 7)             | iProto_roundtrip_ref                                                               |
| - Binder protocol (pg. 10)                         | iProto_binders                                                                     |
| - Smuggle protocol (pg. 10)                        | iProto_smuggle                                                                     |
| - Smuggle reference protocol (pg. 10)              | iProto_smuggle_ref                                                                 |
| - Smuggle recursive protocol (pg. 10)              | iProto_smuggle_ref_rec                                                             |
| Chang and Roberts Examples                         | [multi_actris/examples/leader_election.v](multi_actris/examples/leader_election.v) |
| - `process` program (pg. 12)                       | `process`                                                                          |
| - `init` program (pg. 12)                          | `init`                                                                             |
| - `ring_ref_prog` program (pg. 12)                 | `rle_ref_prog`                                                                     |
| - `ring_prot` protocol (pg. 13)                    | `rle_prot`                                                                         |
| - `init_prot` protocol (pg. 13)                    | `rle_preprot`                                                                      |
| - `ring_ref_prog` protocol network (pg. 13)        | `rle_ref_prog_prot_pool`                                                           |
| - `relay` program (pg. 14)                         | `relay`                                                                            |
| - `ring_del_prog` program (pg. 14)                 | `rle_del_prog`                                                                     |
| - `relay_prot'` protocol (pg. 14)                  | `relay_send_prot`                                                                  |
| - `relay_prot` protocol (pg. 14)                   | `relay_send_preprot`                                                               |
| - `del_prot` protocol (pg. 14)                     | `relay_del_prot`                                                                   |
| - `ring_del_prog` protocol networks (pg. 14)       | `rle_del_prog_prot_pool{,'}`                                                       |
| Two-Buyer Example                                  | [multi_actris/examples/two_buyer.v](multi_actris/examples/two_buyer.v)             |
| Three-Buyer Example                                | [multi_actris/examples/three_buyer.v](multi_actris/examples/three_buyer.v)         |

The reader can observe the `iProto_consistent_take_steps` tactic being used in the verification of the individual protocols.

We recommend the user to step through the proofs (using their Coq IDE of choice), to see how the proof goal changes at each step.

We provide additional insight in how the Multris tactics work in the following reusability guide.

## Reusability Guide

Multris has multiple levels of reusability:

1. Reusing the protocol consistency and meta-theory to apply to new languages
2. Reusing the existing shallow embedding to derive new language consructs
3. Reusing the existing surface-level constructs and rules and verify multiparty message-passing programs

While 1. and 2. require advanced Coq/Iris domain knowledge we make the reusability claim
based on similar prior results based on the Actris framework, that Multris is deeply inspired by,
being applied to the Aneris framework.
More readily available is 3.---reusing Multris to verify new multiparty programs---which we cover in detail in this guide.

Using Multris to verify a multiparty message-passing program consists of three steps.

1. Define the network of protocols
2. Prove consistency of the network using the Multris consistency tactics
3. Verify the program using the Iris/Multris symbolic execution tactics

To become familiar with Multris, we encourage the user to carefully examine the
examples in [multi_actris/examples/basics.v](multi_actris/examples/basics.v),
going through each of the above steps for each them.

Afterwards, we recommend the user to test their familiarity by solving the exercises
in [multi_actris/examples/exercises.v](multi_actris/examples/exercises.v).

The artifact includes solutions to the exercises in [multi_actris/examples/solutions.v](multi_actris/examples/solutions.v),
should one get stuck.

In the following we describe in further detail how to carry out the individual steps of using Multris
for verifying multiparty message-passing programs.

### Define the network of protocols

From inspecting the program, the user should determine and define the network of protocols
via the multiparty dependent separation protocols.

The examples in [multi_actris/examples/basics.v](multi_actris/examples/basics.v)
give some intuition as to how this is done.

### Prove consistency of the network using the Multris consistency tactics

Multris includes a tactic for verifying multiparty protocol consistency:

- `iProto_consistent_take_steps`:
  naively introduces all variables/resources of senders, and eagerly
  instanstiate all variables and resolves resource obligations via IPM framing
  of receivers. If the tactic cannot resolve an obligation by framing it yields
  the remaining proof goal to the user who can manually resolve it.

If the tactic gets stuck the user can try to reach a new state where it can make progress.
This happens most often when one needs to unfold Coq definitions, do case analysis on e.g. booleans,
or use rewrite rules for unfolding recursive definitions.

The examples in [multi_actris/examples/basics.v](multi_actris/examples/basics.v)
give some intuition as to how this is done.

### Verify the program using the Iris/Multris symbolic execution tactics

In order to verify programs using Multris, one can make use of a combination of
[Iris's symbolic execution tactics for HeapLang programs][HeapLang] and
[Multris's symbolic execution tactics for message passing][MultrisProofMode].
The Multris Coq tactics are as follows:

- `wp_new_chan prots with prots_consistent as (c0 .. cn) "Hc0" .. "Hcn"`:
  symbolically execute `new_chan n` and the subsequent `get_chan 0...n`,
  with the protocols `prot`,  and resolve the protocol consistency obligation via `prots_consistent`. 
  The tactic introduces the new channel endpoints as `c0 .. cn` and their
  corresponding endpoint ownership as `Hc0 .. Hcn`.
- `wp_send (t1 .. tn) with "selpat"`: symbolically execute `send c i v` by
  looking up ownership of a send protocol
  `H : c ↣ <(Send,i)> y1 .. yn, MSG v; {{ P }}; prot` in the proof mode context.
  The tactic instantiates the variables `y1 .. yn`
  using the terms `t1 .. tn` and uses `selpat` to prove `P`. If fewer terms
  `t` are given than variables `y`, they will be instantiated using existential
  variables (evars). The tactic will put `H : c ↣ prot` back into the context.
- `wp_recv (x1 .. xn) as "ipat"`: symbolically execute `recv c i` by looking up
  `H : c ↣  <(Recv,i)> y1 .. yn, MSG v; {{ P }}; prot` in the proof mode context.
  The variables `y1 .. yn` are introduced as `x1 .. xn`, and the predicate `P` is
  introduced using the introduction pattern `ipat`. The tactic will put
  `H : c ↣ prot` back into the context.

The above tactics implicitly perform normalization of the protocol `prot` in
the hypothesis `H : c ↣ prot`. For example, `wp_send (40) with [//]` also works if there is a
channel with the protocol `μ rec. <(Send,i) @ (x:Z)> MSG x ; <(Recv,j)> MSG (x+2) ; rec`,
yielding the remaining protocol `<(Recv,j)> MSG (42) ; μ rec. <(Send,i) @ (x:Z)> MSG x ; <(Recv,j)> MSG (x+2) ; rec`.
Concretely, the normalization performs the following actions:

- It unfolds `prot1` into `prot2` if there is an instance of the type class
  `ProtoUnfold prot1 prot2`. When defining a recursive protocol, it is
  useful to define a `ProtoUnfold` instance to obtain automatic unfolding
  of the recursive protocol.
- It re-associates appends (`<++>`), and removes left-identities (`END`) of it.
- It moves appends (`<++>`) into sends (`<!>`), receives (`<?>`), selections
  (`<+>`) and branches (`<&>`).
- It distributes duals (`iProto_dual`) over append (`<++>`).

We omit details about Iris tactics regarding separation logic and symbolic execution, and refer the interested
reader to [ProofMode]: https://gitlab.mpi-sws.org/iris/iris/blob/master/ProofMode.md .
